#include "Dude.h"
#include "OpenGL.h"
#include "Planet.h"
#include "Math.h"
#include "Log.h"
#include "ParticleSystem.h"

#include <assert.h>

Dude::Dude()
{
    m_alive         = true;
    m_direction     = Direction_Clockwise;
    m_state         = State_Moving;
    m_hue           = 0.5f;
    m_cooperation   = 0.0f;
    m_knowledge     = 0.0f;
    m_moveSpeed     = Random(0.00002f, 0.00005f);
    m_genius        = false;
    m_lumber        = 0.0f;
    m_ore           = 0.0f;
}

void Dude::RandomizeDirection()
{
    m_direction = RandomDirection();
}

void Dude::Update(Planet& planet, float deltaTime)
{

    assert(GetIsAlive());

    const float maxSlope = planet.GetMaxImpassableSlope();

    // Check the slope in the direction we're moving.
    float slope = planet.GetLandSlope(m_x, m_direction);

    if (fabs(slope) > maxSlope)
    {
        if (m_state == State_Turning)
        {
            // We're stuck, so just move in the direction of decreasing slope.
            float oppositeSlope = planet.GetLandSlope(m_x, OppositeDirection(m_direction));

            if (oppositeSlope < 0.0f && slope < 0.0f)
            {
                if (oppositeSlope > slope)
                {
                    // Choose the more gradual slope
                    SwitchDirection();
                }
            }
            else if (oppositeSlope < 0.0f)
            {
                // Go down hill.
                SwitchDirection();
            }

            m_state = State_Moving;

        }
        else
        {
            // Switch directions so that next time we start walking back the
            // other way the next update.
            SwitchDirection();
            m_state = State_Turning;
        }
    }
    else
    {
        m_state = State_Moving;
    }
    
    if (m_state == State_Moving)
    {

        float velocity;

        if (m_direction == Direction_Clockwise)
        {
            velocity = m_moveSpeed;
        }
        else
        {
            velocity = -m_moveSpeed;
        }

        // New position is ok, just just move there.
        SetPosition(m_x + velocity * deltaTime);
         
    }

    // Check if we should kill the guy (if we're under water).

    if (planet.GetIsUnderwater(m_x))
    {
        Kill(planet);
    }
 
}

void Dude::Render(const Planet& planet, float time) const
{

    assert(GetIsAlive());

    float offset = cosf(m_x * 300.0f) * 2.0f + 2.0f;

    glColor(GetColor());
    DrawFilledCircle(Vec2(0.0f, 15.0f + offset), 6.0f);

    // If this dude is a genius, put a bobbing exclamation point above his head.

    if (m_genius)
    {

        float offset = cosf(time * 0.01f) * 2.0f + 2.0f;

        DrawFilledCircle(Vec2(0.0f, 26.0f + offset), 3.0f, 8);

        glBegin(GL_QUADS);
        glVertex2f(-2.0f, 30.0f + offset);
        glVertex2f(-3.0f, 40.0f + offset);
        glVertex2f( 3.0f, 40.0f + offset);
        glVertex2f( 2.0f, 30.0f + offset);
        glEnd();
    
    }

}

void Dude::SwitchDirection()
{
    m_direction = OppositeDirection(m_direction);
}

void Dude::SetHue(float hue)
{
    m_hue = hue;
}

float Dude::GetHue() const
{
    return m_hue;
}

void Dude::SetCooperation(float cooperation)
{
    m_cooperation = cooperation;
}

float Dude::GetCooperation() const
{
    return m_cooperation;
}

void Dude::SetFaction(unsigned int faction)
{
    m_faction = faction;
}

unsigned int Dude::GetFaction() const
{
    return m_faction;
}

void Dude::SetKnowledge(float knowledge)
{
    m_knowledge = knowledge;
}

float Dude::GetKnowledge() const
{
    return m_knowledge;
}

void Dude::Kill(Planet& planet)
{

    m_alive = false;

    const float height = 20.0f;
    const float radius = 6.0f;

    ParticleSystem* effect = new ParticleSystem;
    effect->InitializeCircle(GetColor(), planet.GetPoint(m_x, height), radius, 0.1f, 1000.0f);

    planet.AddEffect(effect);

}

Color Dude::GetColor() const
{
    Color color;
    color.SetHSV(m_hue, 0.5f, m_cooperation);
    return color;
}

bool Dude::GetIsAlive() const
{
    return m_alive;
}

bool Dude::GetIsGenius() const
{
    return m_genius;
}

void Dude::SetIsGenius(bool genius)
{
    m_genius = genius;
}

float Dude::GetLumber() const
{
    return m_lumber;
}

void Dude::SetLumber(float lumber)
{
    m_lumber = lumber;
}

float Dude::GetOre() const
{
    return m_ore;
}

void Dude::SetOre(float ore)
{
    m_ore = ore;
}
