#include "Window.h"
#include "Planet.h"
#include "Vec2.h"
#include "Timer.h"
#include "Math.h"
#include "OpenGL.h"
#include "Log.h"

enum GameMode
{
    GameMode_Playing,
    GameMode_Won,
    GameMode_Lost,
};

enum Action
{
    Action_Idle,
    Action_Rotating,
    Action_RaisingGround,
    Action_LoweringGround,
};

GameMode    g_gameMode              = GameMode_Playing;
Action      g_action                = Action_Idle;

Planet*     g_planet                = NULL;

Vec2        g_lastMousePosition;

bool        g_rotating              = false;
float       g_rotateVelocity        = 0.0f;

float       g_currentTime           = 0.0f;
float       g_lastTimeMouseMoved    = 0.0f;

float       g_zoom                  = 400.0f;

void StartGame()
{

    delete g_planet;
    g_planet = NULL;

    g_planet = new Planet;
    g_planet->Initialize();
    
    g_gameMode              = GameMode_Playing;
    g_action                = Action_Idle;
    g_rotating              = false;
    g_rotateVelocity        = 0.0f;
    g_currentTime           = 0.0f;
    g_lastTimeMouseMoved    = 0.0f;
    g_zoom                  = 400.0f;

}

Vec2 GetWorldSpacePoint(const Vec2& screenPoint)
{

    float aspect = g_xSize / g_ySize;

    Vec2 worldPoint;

    worldPoint.x = (g_lastMousePosition.x / g_xSize - 0.5f) * 2.0f * g_zoom * aspect;
    worldPoint.y = (1.0f - g_lastMousePosition.y / g_ySize - 0.5f) * 2.0f * g_zoom;

    return worldPoint;

}

void Render()
{

    glClearColor(1.0f, 1.0f, 1.0f, 1.0);
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

    // Setup the coordinate system so that the origin is in the center of the
    // screen.

    float aspect = g_xSize / g_ySize;

    glMatrixMode(GL_PROJECTION);
    glPushMatrix();
    glLoadIdentity();

    gluOrtho2D(-g_zoom * aspect, g_zoom * aspect, -g_zoom, g_zoom);

    // Render the scene.
    g_planet->Render();

    if (g_gameMode == GameMode_Lost)
    {
        // Draw a big X through the plant to indicate we lost.

        glMatrixMode(GL_MODELVIEW);

        glColor3f(1.0f, 0.25f, 0.25f);

        glPushMatrix();
        glRotatef(-45.0f, 0.0f, 0.0f, 1.0f);

        glBegin(GL_QUADS);
        glVertex2f(-10.0f,  230.0f);
        glVertex2f( 10.0f,  230.0f);
        glVertex2f( 10.0f, -230.0f);
        glVertex2f(-10.0f, -230.0f);
        glEnd();

        glPopMatrix();

        glPushMatrix();
        glRotatef(45.0f, 0.0f, 0.0f, 1.0f);

        glBegin(GL_QUADS);
        glVertex2f(-10.0f,  230.0f);
        glVertex2f( 10.0f,  230.0f);
        glVertex2f( 10.0f, -230.0f);
        glVertex2f(-10.0f, -230.0f);
        glEnd();

        glPopMatrix();

    }
    else if (g_gameMode == GameMode_Won)
    {

        // Draw a big check through the plant to indicate we won.

        glMatrixMode(GL_MODELVIEW);

        glColor3f(0.25f, 1.0f, 0.25f);

        glPushMatrix();
        glRotatef(-45.0f, 0.0f, 0.0f, 1.0f);

        glBegin(GL_QUADS);
        glVertex2f(-10.0f,  230.0f);
        glVertex2f( 10.0f,  230.0f);
        glVertex2f( 10.0f, -230.0f);
        glVertex2f(-10.0f, -230.0f);
        glEnd();

        glPopMatrix();

        glPushMatrix();
        glTranslatef(-191.0f, -121.0f, 0.0f);
        glRotatef(45.0f, 0.0f, 0.0f, 1.0f);

        glBegin(GL_QUADS);
        glVertex2f(-10.0f,  50.0f);
        glVertex2f( 10.0f,  50.0f);
        glVertex2f( 10.0f, -50.0f);
        glVertex2f(-10.0f, -50.0f);
        glEnd();

        glPopMatrix();

    }

    glMatrixMode(GL_PROJECTION);
    glPopMatrix();

    // Render screen space elements

    glMatrixMode(GL_PROJECTION);
    glPushMatrix();
    glLoadIdentity();
    gluOrtho2D(0.0, g_xSize, g_ySize, 0.0f);


    // Check if we need to render the popup (do so after the mouse hasn't moved
    // for a certain amount of time).

    float popupWaitTime = 300.0f;

    if (g_currentTime - g_lastTimeMouseMoved > popupWaitTime)
    {
        if (g_action == Action_Idle)
        {
            Vec2 popupLocation = GetWorldSpacePoint(g_lastMousePosition); 
            g_planet->RenderPopup(g_lastMousePosition, popupLocation);
        }
    }

    glMatrixMode(GL_PROJECTION);
    glPopMatrix();

}

void OnMouseMove(int button[2])
{

    Vec2 mousePosition = GetMousePosition();

    if (g_action == Action_Rotating)
    {

        Vec2 center;

        center.x = g_xSize * 0.5f;
        center.y = g_ySize * 0.5f;

        Vec2 v = (g_lastMousePosition - center).GetUnit();
        Vec2 w = (mousePosition - center).GetUnit();

        g_rotateVelocity += (v.x * w.y - v.y * w.x) * 0.02f;

    }

    g_lastTimeMouseMoved    = g_currentTime;
    g_lastMousePosition     = mousePosition;

}

void OnMouseDown(int button)
{

    Vec2 mouse = GetMousePosition();

    Vec2 worldPoint = GetWorldSpacePoint(mouse);
    float distance = worldPoint.GetLength();
    
    if (distance > 300.0f)
    {
        if (button == 0)
        {
            g_action = Action_Rotating;
        }
    }
    else
    {
        if (button == 0)
        {
            g_action = Action_RaisingGround;
        }
        else if (button == 1)
        {
            g_action = Action_LoweringGround;
        }
    }

    g_lastMousePosition = mouse;

    if (g_gameMode == GameMode_Lost)
    {
        // Restart the game.
        StartGame();
    }

}

void OnMouseUp(int button)
{
    g_action = Action_Idle;
}

void ProcessInput(float deltaTime)
{

    Vec2 center;

    center.x = g_xSize * 0.5f;
    center.y = g_ySize * 0.5f;

    Vec2 mouse = GetMousePosition();

    float angle = atan2(center.y - mouse.y, mouse.x - center.x) + g_planet->GetAngle();
    float x = angle / (2.0f * pi);

    float raiseAmount = 0.1f * deltaTime;
    float raiseRadius = 0.05f;

    if (g_action == Action_RaisingGround)
    {
        // Raise the land.
        g_planet->RaiseLand(x, raiseRadius, raiseAmount);
    }
    else if (g_action == Action_LoweringGround)
    {
        // Lower the land.
        g_planet->RaiseLand(x, raiseRadius, -raiseAmount);
    }

}

void Update(float deltaTime)
{
    
    // Update the rotation.
    float angle = g_planet->GetAngle() + g_rotateVelocity * deltaTime;
    g_planet->SetAngle(angle);

    // Dampen the rotation of the planet.

    float friction = 0.00002f;
    
    if (g_rotateVelocity > 0.0f)
    {
        g_rotateVelocity = max(g_rotateVelocity - friction * deltaTime, 0.0f);
    }
    else if (g_rotateVelocity < 0.0f)
    {
        g_rotateVelocity = min(g_rotateVelocity + friction * deltaTime, 0.0f);
    }
    
    // Update the planet.
    g_planet->Update(deltaTime);
        
    if (g_gameMode == GameMode_Playing)
    {

        // Check to see if we lost the game.
        if (g_planet->GetNumDudes() == 0)
        {
            g_gameMode = GameMode_Lost;
        }
        else if (g_planet->GetHasWon())
        {
            g_gameMode = GameMode_Won;
        }

    }

}

int WINAPI WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPSTR lpCmdLine, int nCmdShow)
{

    const float maxDeltaTime = 0.1f;

	if (!CreateGLWindow(L"Shift - Max McGuire", 800, 600, 32, FALSE))
	{
		return 0;
	}

    // Setup alpha blending.
    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
    glEnable(GL_BLEND);

    Timer frameTimer;

    StartGame();

	while (ProcessMessages())
	{

        frameTimer.Update();
        float deltaTime = frameTimer.GetElapsedTime();

        g_currentTime += deltaTime;

        // Process the input.
        ProcessInput(deltaTime);

        // Update the game at a fixed time step.
        Update(deltaTime);

        // Render
        Render();

        //Log::Get().Render();

        SwapBuffers();

	}

    delete g_planet;

	KillGLWindow();
	return 0;

}
