#include "Planet.h"
#include "OpenGL.h"
#include "Noise.h"
#include "Math.h"
#include "Vec2.h"
#include "Dude.h"
#include "Effect.h"
#include "Plant.h"
#include "Alert.h"
#include "Nuke.h"
#include "ParticleSystem.h"
#include "Sparkle.h"
#include "Meteor.h"
#include "Meteorite.h"
#include "Building.h"
#include "Rocket.h"

#include <assert.h>
#include <algorithm>

const float Planet::s_maxImpassableSlope = 25.0f;

template <typename T>
void Clear(std::vector<T*>& v)
{
    for (unsigned int i= 0; i < v.size(); ++i)
    {
        delete v[i];
    }
    v.clear();
}

template <typename T>
struct ObjectSort
{
    bool operator()(const T* obj1, const T* obj2)
    {
        return obj1->GetPosition() < obj2->GetPosition();
    }
};


bool Planet::Faction::ContainsPosition(float x) const
{
    // Put x in the same range as the faction limits so that we properly handle
    // factions that wrap around the 0 point.
    x = Wrap(x, min, min + 1.0f);
    return x >= min && x < max;
}

Planet::~Planet()
{
    Clear(m_dudes);
    Clear(m_effects);
    Clear(m_plants);
    Clear(m_meteors);
    Clear(m_meteorites);
    Clear(m_buildings);
    Clear(m_rockets);
}

void Planet::Initialize()
{

    m_won = false;

    m_popupFont.Initialize(L"Tahoma");

    const float variation   = 50.0f;
    const float baseHeight  = 200.0f;

    for (int i = 0; i < s_numLandSamples; ++i)
    {
        m_landHeight[i] = baseHeight;
    }

    m_maxDudes = 100;
    Populate(m_maxDudes);

    m_waterLevel    = 180.0f;
    m_waterTime     = 0.0f;

    m_angle         = 0.0f;

    m_sunCenter     = Vec2(450.0f, 230.0f);
    m_sunRadius     = 100.0f;

    m_plantSpawnTimer   = 0.0f;
    m_dudeSpawnTimer    = 0.0f;
    m_meteorSpawnTimer  = 0.0f;

    UpdateLandSlope();

}

void Planet::SetAngle(float angle)
{
    m_angle = angle;
}

float Planet::GetAngle() const
{
    return m_angle;
}

void Planet::Render() const
{

    glMatrixMode(GL_MODELVIEW);
    glPushMatrix();
    
    glRotatef(-GetDegrees(m_angle), 0.0f, 0.0f, 1.0f);

    RenderFactions();

    // Eliminate the rotation matrix.
    glPushMatrix();
    glLoadIdentity();
    RenderMeteors();
    glPopMatrix();

    RenderMeteorites();

    RenderPlants();
    RenderBuildings();
    RenderRockets();
    RenderDudes();
    RenderWater();


    glColor3f(0.0f, 0.0f, 0.0f);
    glBegin(GL_TRIANGLES);

    const float deltaAngle = 2.0f * pi / (s_numLandSamples - 1.0f);

    Vec2 point1;
    Vec2 point2;

    point1.x = cosf(-deltaAngle) * m_landHeight[s_numLandSamples];
    point1.y = sinf(-deltaAngle) * m_landHeight[s_numLandSamples];

    for (int i = 0; i < s_numLandSamples; ++i)
    {

        float angle = i * deltaAngle;

        point2.x = cosf(angle) * m_landHeight[i];
        point2.y = sinf(angle) * m_landHeight[i];

        glPoint(Vec2(0.0f, 0.0f));
        glPoint(point1);
        glPoint(point2);

        point1 = point2;

    }

    glEnd();

    // Draw a white ring to give the planet a more round appearance.
    glColor3f(1.0f, 1.0f, 1.0f);
    DrawCircle(Vec2(0.0f, 0.0f), GetSurfaceHeight(), 128);

    // Render the effects.
    for (unsigned int i = 0; i < m_effects.size(); ++i)
    {
        m_effects[i]->Render(*this);
    }

    /*
    // Indicate the sunny side.

    glColor3f(1.0f, 1.0f, 0.0f);
    glBegin(GL_LINES);

    glVertex2f(0.0f, 0.0f);q
    glVertex2f(cosf(m_angle + pi / 2.0f) * 400, sinf(m_angle + pi / 2.0f) * 400);
    
    glVertex2f(0.0f, 0.0f);
    glVertex2f(cosf(m_angle + pi * 1.75f) * 400, sinf(m_angle + pi * 1.75f) * 400);

    glEnd();
    */

    glMatrixMode(GL_MODELVIEW);
    glPopMatrix();

    RenderSun();

}

void Planet::RenderDudes() const
{
    for (unsigned int i = 0; i < m_dudes.size(); ++i)
    {
        RenderThing(m_dudes[i]);
    }
}

void Planet::RenderBuildings() const
{
    for (unsigned int i = 0; i < m_buildings.size(); ++i)
    {
        RenderThing(m_buildings[i]);
    }
}

void Planet::RenderRockets() const
{
    for (unsigned int i = 0; i < m_rockets.size(); ++i)
    {
        RenderThing(m_rockets[i]);
    }
}

void Planet::RenderThing(const Thing* thing) const
{

    // Setup the local transform.

    float x         = thing->GetPosition();
    float angle     = x * 2.0f * pi;
    float radius    = GetLandHeight(x);

    float dx = cosf(angle);
    float dy = sinf(angle);

    GLfloat matrix[16];

    matrix[0]  = -dy;
    matrix[1]  = dx;
    matrix[2]  = 0.0f;
    matrix[3]  = 0.0f;

    matrix[4]  = dx;
    matrix[5]  = dy;
    matrix[6]  = 0.0f;
    matrix[7]  = 0.0f;

    matrix[8]  = 0.0f;
    matrix[9]  = 0.0f;
    matrix[10] = 1.0f;
    matrix[11] = 0.0f;

    matrix[12] = cosf(angle) * radius;
    matrix[13] = sinf(angle) * radius;
    matrix[14] = 0.0f;
    matrix[15] = 1.0f;

    glMatrixMode(GL_MODELVIEW);
    glPushMatrix();
    glMultMatrixf(matrix);

    thing->Render(*this, m_waterTime);

    glPopMatrix();


}

void Planet::RenderPlants() const
{
    for (unsigned int i = 0; i < m_plants.size(); ++i)
    {
        RenderThing(m_plants[i]);
    }
}

void Planet::RenderFactions() const
{

    for (unsigned int i = 0; i < m_factions.size(); ++i)
    {

        const Faction& faction = m_factions[i];

        if (faction.numDudes > 0)
        {

            float minAngle = m_factions[i].min * 2.0f * pi;
            float maxAngle = m_factions[i].max * 2.0f * pi;

            // Draw a pie wedge.

            glColor(GetFactionColor(i));
            glBegin(GL_TRIANGLE_FAN);
            glPoint(Vec2(0.0f, 0.0f));

            for (float angle = minAngle; angle < maxAngle; angle += 0.01f)
            {
                glPoint(Vec2(cosf(angle) * 300.0f, sinf(angle) * 300.0f));
            }

            glPoint(Vec2(cosf(maxAngle) * 300.0f, sinf(maxAngle) * 300.0f));
            glEnd();

            if (fabsf(faction.min - faction.max) != 1.0f)
            {

                glLineWidth(2.0f);
                glBegin(GL_LINES);
                glColor3f(0.0f, 0.0f, 0.0f);

                glPoint(Vec2(0.0f, 0.0f));
                glPoint(Vec2(cosf(minAngle) * 300.0f, sinf(minAngle) * 300.0f));

                glPoint(Vec2(0.0f, 0.0f));
                glPoint(Vec2(cosf(maxAngle) * 300.0f, sinf(maxAngle) * 300.0f));

                glEnd();

            }

        }

    }

}

void Planet::RenderWater() const
{

    const int numWaterSamples = 1000;
    const float deltaAngle = 2.0f * pi / (numWaterSamples - 1.0f);

    // Draw the white background for underwater (i.e. erase what's behind the waves).

    glColor3f(1.0f, 1.0f, 1.0f);
    glBegin(GL_TRIANGLES);

    Vec2 point1;
    Vec2 point2;

    float waveHeight = GetWaveHeight(-deltaAngle);

    point1.x = cosf(-deltaAngle) * (m_waterLevel + waveHeight);
    point1.y = sinf(-deltaAngle) * (m_waterLevel + waveHeight);

    for (int i = 0; i < numWaterSamples; ++i)
    {

        float angle = i * deltaAngle;
        float waveHeight = GetWaveHeight(angle);

        point2.x = cosf(angle) * (m_waterLevel + waveHeight);
        point2.y = sinf(angle) * (m_waterLevel + waveHeight);

        glPoint(Vec2(0.0f, 0.0f));
        glPoint(point1);
        glPoint(point2);

        point1 = point2;

    }

    glEnd();

    // Draw a black outline for the top of the waves.

    glColor3f(0.0f, 0.0f, 0.0f);

    glLineWidth(2.0f);
    glBegin(GL_LINE_STRIP);

    for (int i = 0; i < numWaterSamples; ++i)
    {

        float angle = i * deltaAngle;
        float waveHeight = GetWaveHeight(angle);

        Vec2 point;

        point.x = cosf(angle) * (m_waterLevel + waveHeight);
        point.y = sinf(angle) * (m_waterLevel + waveHeight);

        glPoint(point);

    }

    glEnd();

}

void Planet::Update(float deltaTime)
{

    m_meteorSpawnTimer -= deltaTime;
    SpawnMeteors();

    m_plantSpawnTimer -= deltaTime;
    SpawnPlants();

    // Sort the dudes by positions to make other calculations easier.
    std::sort(m_dudes.begin(), m_dudes.end(), ObjectSort<Dude>());
    std::sort(m_plants.begin(), m_plants.end(), ObjectSort<Plant>());

    // Rebuild the factions for the planet.
    BuildFactions();
    
    // Update the dudes and factions based on the new configuration.
    UpdateDudeFactions();
    UpdateFactionStats();

    // For every genius/builing pair, increase knowledge of the faction.
    for (unsigned int i = 0; i < m_factions.size(); ++i)
    {
        int geniusBuildingPairs = min(m_factions[i].numGeniuses, m_factions[i].numBuildings);
        m_factions[i].knowledge += m_factions[i].cooperation * geniusBuildingPairs * deltaTime * 0.1f;
    }

    ApplyFactions(deltaTime);

    m_waterTime += deltaTime;
    UpdateDudes(deltaTime);

    // Spawn dudes last since we do this based on factions.
    m_dudeSpawnTimer += deltaTime;
    SpawnDudes();

    MakeFactionPurchases();

    UpdatePlants(deltaTime);
    UpdateMeteors(deltaTime);   
    UpdateMeteorites(deltaTime);
    UpdateEffects(deltaTime);
    UpdateBuildings(deltaTime);

    ProcessFactionDiplomacy();

    // Check for win.

    const int dudesPerRocket = 20;

    unsigned int goodFactions = 0;

    for (unsigned int i = 0; i < m_factions.size(); ++i)
    {
        if (m_factions[i].numRockets * dudesPerRocket > m_factions[i].numDudes)
        {
            ++goodFactions;
        }
    }

    m_won = goodFactions == m_factions.size();

}

void Planet::UpdateDudes(float deltaTime)
{

    for (unsigned int i = 0; i < m_dudes.size(); /* */)
    {

        m_dudes[i]->Update(*this, deltaTime);

        // Remove the dead guy.c

        if (!m_dudes[i]->GetIsAlive())
        {
            delete m_dudes[i];
            m_dudes[i] = m_dudes[m_dudes.size() - 1];
            m_dudes.resize(m_dudes.size() - 1);
        }
        else
        {
            ++i;
        }

    }

}

void Planet::UpdateBuildings(float deltaTime)
{

    for (unsigned int i = 0; i < m_buildings.size(); /* */)
    {

        m_buildings[i]->Update(*this, deltaTime);

        // Remove the dead guy.

        if (!m_buildings[i]->GetIsAlive())
        {
            delete m_buildings[i];
            m_buildings[i] = m_buildings[m_buildings.size() - 1];
            m_buildings.resize(m_buildings.size() - 1);
        }
        else
        {
            ++i;
        }

    }

}


void Planet::UpdatePlants(float deltaTime)
{

    for (unsigned int i = 0; i < m_plants.size(); /* */)
    {

        m_plants[i]->Update(*this, deltaTime);

        // Remove the dead guy.

        if (!m_plants[i]->GetIsAlive())
        {
            delete m_plants[i];
            m_plants[i] = m_plants[m_plants.size() - 1];
            m_plants.resize(m_plants.size() - 1);
        }
        else
        {
            ++i;
        }

    }

}

void Planet::UpdateEffects(float deltaTime)
{

    for (unsigned int i = 0; i < m_effects.size(); /* */)
    {

        m_effects[i]->Update(*this, deltaTime);

        // Remove the finished particle effect.

        if (!m_effects[i]->GetIsAlive())
        {
            delete m_effects[i];
            m_effects[i] = m_effects[m_effects.size() - 1];
            m_effects.resize(m_effects.size() - 1);
        }
        else
        {
            ++i;
        }

    }

}

void Planet::RaiseLand(float x, float radius, float amount)
{

    const float minLandHeight = m_waterLevel - 20.0f;
    const float maxLandHeight = 300.0f;

    float min = x - radius;
    float max = x + radius;

    int imin = static_cast<int>(floorf(min * s_numLandSamples));
    int imax = static_cast<int>(ceilf(max * s_numLandSamples));

    for (int i = imin; i < imax; ++i)
    {

        float attenuate = powf(Bump((static_cast<float>(i) / s_numLandSamples - x) / radius), 5.0f);
        
        int wi = Wrap(i, 0, s_numLandSamples);
        m_landHeight[wi] = min(max(m_landHeight[wi] + amount * attenuate, minLandHeight), maxLandHeight);

    }

    UpdateLandSlope();

}

float Planet::GetLandHeight(float x) const
{

    float sx = x * s_numLandSamples;

    int min = Wrap(static_cast<int>(floor(sx)), 0, s_numLandSamples);
    int max = Wrap(static_cast<int>(ceil(sx)), 0, s_numLandSamples);

    float dx = sx - floor(sx);

    // Linear filtering.
    return Lerp(m_landHeight[min], m_landHeight[max], dx);

}

float Planet::GetLandSlope(float x, Direction direction) const
{

    float sx = x * s_numLandSamples;

    int min = Wrap(static_cast<int>(floor(sx)), 0, s_numLandSamples);
    int max = Wrap(static_cast<int>(ceil(sx)), 0, s_numLandSamples);

    float dx = sx - floor(sx);

    // Linear filtering.
    return Lerp(m_landSlope[min][direction], m_landSlope[max][direction], dx);

}


float Planet::GetLandSlope(float x, float direction) const
{

    float min = x;
    float max = x + direction;

    int imin = static_cast<int>(floorf(min * s_numLandSamples));
    int imax = static_cast<int>(ceilf(max * s_numLandSamples));

    if (imin > imax)
    {
        Swap(imin, imax);
    }

    float xh = GetLandHeight(x);

    float dmax = 0.0f;
    float dmin = 0.0f;

    for (int i = imin; i < imax; ++i)
    {

        float y = static_cast<float>(i) / s_numLandSamples;

        int wi = Wrap(i, 0, s_numLandSamples);
        float yh = m_landHeight[wi];

        // A bit hacky, but just consider the slope real steep if this goes under water.
        // The makes water act as a faction boundary.
        if (yh < m_waterLevel)
        {
            return 10000.0f;
        }

        float d = direction * (yh - xh) / (y - x);

        dmin = min(d, dmin);
        dmax = max(d, dmax);
        
    }

    if (-dmin > dmax)
    {
        return dmin;
    }
    else
    {
        return dmax;
    }

}

Vec2 Planet::GetPoint(float x, float height) const
{

    float angle = x * 2.0f * pi;

    float landHeight    = GetLandHeight(x);
    float totalHeight   = landHeight + height;

    return Vec2(cosf(angle) * totalHeight, sinf(angle) * totalHeight);
    
}

float Planet::Fractal(float x)
{

    const int octaves = 8;

    float result    = 0.0f;

    float scale     = 1.0f;
    float factor    = 1.0f;

    for (int i = 0; i < octaves; ++i)
    {
        result += noise1(x * scale) * factor;
        scale   *= 2.0f;
        factor  *= 0.5f;
    }

    return result;

}

void Planet::Populate(unsigned int numDudes)
{

    Clear(m_dudes);
    m_dudes.resize(numDudes);

    for (unsigned int i = 0; i < numDudes; ++i)
    {

        Dude* dude = new Dude;
        dude->SetPosition(Random(0.0f, 1.0f));
        dude->RandomizeDirection();
        
        m_dudes[i] = dude;
        
    }

}

void Planet::BuildFactions()
{
    
    m_factions.resize(0);

    // Find the start of our first faction.

    int firstFactionSample = 0;

    while (!GetIsImpassableBoundary(firstFactionSample) && firstFactionSample < s_numLandSamples)
    {
        ++firstFactionSample;
    }

    // Build the factions

    const float minFactionSize  = 0.01f;
    float factionMin = firstFactionSample / static_cast<float>(s_numLandSamples);

    for (int i = firstFactionSample; i < s_numLandSamples + firstFactionSample; ++i)
    {

        int wi = Wrap(i, 0, s_numLandSamples);

        // Check if this spot has a steep enough slope to divide groups of dudes.

        float factionMax = i / static_cast<float>(s_numLandSamples);

        if (GetIsImpassableBoundary(wi) && factionMax - factionMin > minFactionSize)
        {

            // Create a new faction.

            Faction faction;
            faction.min = factionMin;
            faction.max = factionMax;

            m_factions.push_back(faction);

            factionMin = faction.max;

        }
       
    }

    // Close the last faction.

    Faction faction;
    faction.min = factionMin;
    faction.max = static_cast<float>((s_numLandSamples + firstFactionSample)) / s_numLandSamples;

    m_factions.push_back(faction);

    // Update the hue for the factions.
    for (unsigned int i = 0; i < m_factions.size(); ++i)
    {
        // We offset so that when we have one big faction it's not mauve.
        m_factions[i].hue = Wrap((m_factions[i].min + m_factions[i].max) * 0.5f + 0.15f, 0.0, 1.0f);
    }

}

void Planet::UpdateLandSlope()
{

    const float lookAhead = 0.01f;

    for (int i = 0; i < s_numLandSamples; ++i)
    {
        float x = static_cast<float>(i) / s_numLandSamples;
        m_landSlope[i][Direction_Clockwise]         = GetLandSlope(x, lookAhead);
        m_landSlope[i][Direction_CounterClockwise]  = GetLandSlope(x, -lookAhead);
    }

}

void Planet::UpdateDudeFactions()
{

    unsigned int currentFaction = 0;

    for (unsigned int i = 0; i < m_dudes.size(); ++i)
    {
        
        Dude* dude = m_dudes[i];

        // Advance to the faction for this dude.
        while (!m_factions[currentFaction].ContainsPosition(dude->GetPosition()))
        {
            currentFaction = (currentFaction + 1) % m_factions.size();
        }

        dude->SetFaction(currentFaction);

    }

}

void Planet::UpdateFactionResourceCounts()
{

    for (unsigned int i = 0; i < m_factions.size(); ++i)
    {
        m_factions[i].numPlants     = 0;
        m_factions[i].numOre        = 0;
        m_factions[i].numBuildings  = 0;
        m_factions[i].numRockets    = 0;
    }

    unsigned int currentFaction = 0;

    for (unsigned int i = 0; i < m_plants.size(); ++i)
    {
        
        Plant* plant = m_plants[i];

        // Advance to the faction for this dude.
        while (!m_factions[currentFaction].ContainsPosition(plant->GetPosition()))
        {
            currentFaction = (currentFaction + 1) % m_factions.size();
        }

        ++m_factions[currentFaction].numPlants;

    }

    currentFaction = 0;

    for (unsigned int i = 0; i < m_meteorites.size(); ++i)
    {
        
        Meteorite* meteorite = m_meteorites[i];

        // Advance to the faction for this dude.
        while (!m_factions[currentFaction].ContainsPosition(meteorite->GetPosition()))
        {
            currentFaction = (currentFaction + 1) % m_factions.size();
        }

        ++m_factions[currentFaction].numOre;

    }

    currentFaction = 0;

    for (unsigned int i = 0; i < m_buildings.size(); ++i)
    {
        
        Building* building = m_buildings[i];

        // Advance to the faction for this dude.
        while (!m_factions[currentFaction].ContainsPosition(building->GetPosition()))
        {
            currentFaction = (currentFaction + 1) % m_factions.size();
        }

        ++m_factions[currentFaction].numBuildings;

    }

    currentFaction = 0;

    for (unsigned int i = 0; i < m_rockets.size(); ++i)
    {
        
        Rocket* rocket = m_rockets[i];

        // Advance to the faction for this dude.
        while (!m_factions[currentFaction].ContainsPosition(rocket->GetPosition()))
        {
            currentFaction = (currentFaction + 1) % m_factions.size();
        }

        ++m_factions[currentFaction].numRockets;

    }

}

void Planet::ApplyFactions(float deltaTime)
{

    for (unsigned int i = 0; i < m_dudes.size(); ++i)
    {
        
        Dude* dude = m_dudes[i];
        const Faction& faction = m_factions[dude->GetFaction()];

        // Update the dudes hue based on the hue for the faction.
        dude->SetHue(Interpolate(dude->GetHue(), faction.hue, deltaTime * 0.0001f));

        // Update the dudes knowledge based on the knowledge for the faction.
        dude->SetKnowledge(Interpolate(dude->GetKnowledge(), faction.knowledge, deltaTime * 0.001f));

        // Update the cooperation for the dudes based on the inverse of the
        // number of dudes in the faction (small factions are more cooperative).
        dude->SetCooperation(Interpolate(dude->GetCooperation(), faction.cooperation, deltaTime * 0.0001f));

    }

}

void Planet::UpdateFactionStats()
{

    for (unsigned int i = 0; i < m_factions.size(); ++i)
    {
        m_factions[i].numDudes      = 0;
        m_factions[i].numGeniuses   = 0;
        m_factions[i].knowledge     = 0.0f;
    }

    UpdateFactionResourceCounts();

    
    for (unsigned int i = 0; i < m_dudes.size(); ++i)
    {

        Dude* dude = m_dudes[i];
        
        unsigned int faction = dude->GetFaction();
        assert(faction < m_factions.size());

        ++m_factions[faction].numDudes;

        if (dude->GetIsGenius())
        {
            ++m_factions[faction].numGeniuses;
        }

        m_factions[faction].knowledge += dude->GetKnowledge();

    }

    // Compute the averages for the faction stats.
    for (unsigned int i = 0; i < m_factions.size(); ++i)
    {
        m_factions[i].knowledge /= m_factions[i].numDudes;
    }

    for (unsigned int i = 0; i < m_factions.size(); ++i)
    {
        
        Faction& faction = m_factions[i];
        faction.cooperation  = 1.0f - Clamp(static_cast<float>(faction.numDudes) / s_factionAnarchySize, 0.0f, 1.0f);;
    
        if (faction.numDudes > 0)
        {
            // The more plants, the better as long as people are cooperating.
            float area = (faction.max - faction.min);
            faction.reproduction = faction.cooperation * faction.numPlants * area;
        }
        else
        {
            // No dudes, no reproduction!
            faction.reproduction = 0.0f;
        }
    
    }

}

bool Planet::GetIsImpassableBoundary(int i) const
{

    // Check if the slope is too steep (i.e. mountains).

    float slope1 = m_landSlope[i][Direction_Clockwise];
    float slope2 = m_landSlope[i][Direction_Clockwise];

    float slope = std::max<float>(fabsf(slope1), fabsf(slope2));

    return slope > s_maxImpassableSlope;

}

float Planet::GetWaveHeight(float angle) const
{
    return 3.0f * cosf(angle * 10.0f + m_waterTime * 0.01f) * sinf(angle * 40.0f + m_waterTime * 0.001f);
}

float Planet::GetMaxImpassableSlope() const
{
    return s_maxImpassableSlope;
}

void Planet::AddEffect(Effect* effect)
{
    m_effects.push_back(effect);
}

bool Planet::GetIsUnderwater(float x) const
{
    return GetLandHeight(x) < m_waterLevel;
}

float Planet::GetSurfaceHeight() const
{
    return m_waterLevel - 4.0f;
}

void Planet::RenderSun() const
{

    glLineWidth(2.0f);

    // Draw the core of the sun.
    glColor3f(0.0f, 0.0f, 0.0f);
    DrawCircle(m_sunCenter, m_sunRadius, 64);

    // Draw the rings emmanating from the sun.

    const int numRings = 4;

    for (int i = 0; i < numRings; ++i)
    {

        float f = fmodf(m_waterTime * 0.01f + i * 10.0f, 40.0f) / 40.0f;

        glColor3f(f, f, f);
        DrawCircle(m_sunCenter, m_sunRadius + f * 40.0f, 64);
    
    }

}

unsigned int Planet::GetNumDudes() const
{
    return static_cast<unsigned int>(m_dudes.size());
}

void Planet::RenderPopup(const Vec2& screenLocation, const Vec2& worldLocation) const
{

    float distanceFromPlanet = worldLocation.GetLength();
        
    if ((worldLocation - m_sunCenter).GetLength() < m_sunRadius)
    {
        RenderPopup(screenLocation, "Sol", 40.0f, m_popupFont.GetHeight());
    }
    else if (distanceFromPlanet < GetSurfaceHeight())
    {
        RenderPopup(screenLocation, "Terra", 50.0f, m_popupFont.GetHeight());
    }
    else if (distanceFromPlanet < 300.0f)
    {

        // Faction popup.

        float angle = atan2(worldLocation.y, worldLocation.x) + m_angle;
        float x = angle / (2.0f * pi);

        for (unsigned int i = 0; i < m_factions.size(); ++i)
        {
            const Faction& faction = m_factions[i];
            if (faction.ContainsPosition(x) && faction.numDudes > 0)
            {
                
                char message[1024];
                sprintf(message, "Faction\n%d dudes\n%d lumber\n%d ore\n%0.0f%% cooperation\n%0.2f reproduction\n%0.2f knowledge",
                    faction.numDudes, faction.numPlants, faction.numOre, faction.cooperation * 100.0f, faction.reproduction, faction.knowledge);
    
                RenderPopup(screenLocation, message, 200.0f, m_popupFont.GetHeight() * 8);
                break;

            }
        }

    }

}

void Planet::RenderPopup(const Vec2& screenLocation, const char* message, float xSize, float ySize) const
{

    xSize += 5;
    ySize += 5;

    const Vec2 popupOffset = Vec2(10.0f, 1.0f);

    glBegin(GL_QUADS);
    glColor3f(1.0f, 1.0f, 1.0f);
    glPoint(screenLocation + popupOffset);
    glPoint(screenLocation + popupOffset + Vec2(xSize, 0.0f));
    glPoint(screenLocation + popupOffset + Vec2(xSize, ySize));
    glPoint(screenLocation + popupOffset + Vec2(0.0f,  ySize));
    glEnd();

    glLineWidth(1.0f);
    glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);    

    glBegin(GL_QUADS);
    glColor3f(0.0f, 0.0f, 0.0f);
    glPoint(screenLocation + popupOffset);
    glPoint(screenLocation + popupOffset + Vec2(xSize, 0.0f));
    glPoint(screenLocation + popupOffset + Vec2(xSize, ySize));
    glPoint(screenLocation + popupOffset + Vec2(0.0f,  ySize));
    glEnd();

    glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);  

    glColor3f(0.0f, 0.0f, 0.0f);

    char* messageDup = _strdup(message);
    char* line = messageDup;
    
    for (int i = 0; line != NULL; ++i)
    {

        char* end = strchr(line, '\n');

        if (end != NULL)
        {
            end[0] = 0;
            ++end;
        }

        glRasterPos2f(screenLocation.x + popupOffset.x + 5.0f, screenLocation.y + popupOffset.y + m_popupFont.GetHeight() * (i + 1));
        m_popupFont.Print(line);
        
        line = end;
    
    }

    free(messageDup);
    messageDup = NULL;

}

void Planet::SpawnPlants()
{

    // These are roughly based on what part of the planet is visible to the sun.
    float minAngle = m_angle + pi * 0.50f;
    float maxAngle = m_angle + pi * 1.75f;

    float min = minAngle / (2.0f * pi);
    float max = maxAngle / (2.0f * pi);

    float plantSpawnInterval = 1000.0f;
    
    while (m_plantSpawnTimer < 0.0f)
    {

        // Choose a random point on the sunny side.
        float x = Wrap(Random(max, min + 1.0f), 0.0f, 1.0f);
        
        // Check to see if there's alreay a plant near here.
        if (GetCanSupportPlants(x) && !GetIsPlantNear(x, 0.01f))
        {
            Plant* plant = new Plant;
            plant->SetPosition(x);
            m_plants.push_back(plant);
        }

        m_plantSpawnTimer += Random(0.0f, plantSpawnInterval);
    
    }

}

bool Planet::GetCanSupportPlants(float x) const
{

    if (GetIsUnderwater(x))
    {
        // Don't let things grow underwater.
        return false;
    }

    // Check to see if there is water nearby.

    float radius = 0.08f;

    float min = x - radius;
    float max = x + radius;

    int imin = static_cast<int>(floorf(min * s_numLandSamples));
    int imax = static_cast<int>(ceilf(max * s_numLandSamples));

    for (int i = imin; i < imax; ++i)
    {
        int wi = Wrap(i, 0, s_numLandSamples);
        if (m_landHeight[wi] < m_waterLevel)
        {
            return true;
        }
    }

    return false;

}

bool Planet::GetIsSunny(float x) const
{
    
    float minAngle = m_angle + pi * 0.50f;
    float maxAngle = m_angle + pi * 1.75f;

    float min = minAngle / (2.0f * pi);
    float max = maxAngle / (2.0f * pi);
    
    x = Wrap(x, max, max + 1.0f);
    return x >= max && x < min + 1.0f;

}

bool Planet::GetIsPlantNear(float x, float radius) const
{

    // TODO
    // This doesn't properly handle the boundary case (although probably not
    // terribly important).

    for (unsigned int i = 0; i < m_plants.size(); ++i)
    {
        if (fabsf(m_plants[i]->GetPosition() - x) < radius)
        {
            return true;
        }
    }

    return false;

}

bool Planet::GetIsBuildingNear(float x, float radius) const
{

    // TODO
    // This doesn't properly handle the boundary case (although probably not
    // terribly important).

    for (unsigned int i = 0; i < m_buildings.size(); ++i)
    {
        if (fabsf(m_buildings[i]->GetPosition() - x) < radius)
        {
            return true;
        }
    }

    return false;

}


void Planet::SpawnDudes()
{

    float dudeSpawnInterval = 10000.0f;
    
    while (m_dudeSpawnTimer > dudeSpawnInterval)
    {

        // If the planet can support more dudes, go for it.
        if (m_dudes.size() < m_maxDudes)
        {

            unsigned int faction = GetRandomizedFactionToPopulate();

            if (faction != -1)
            {

                Dude* dude = new Dude;
                dude->SetPosition(Random(m_factions[faction].min, m_factions[faction].max));
                dude->RandomizeDirection();

                // Randomly make this guy a genius with low probability.
                dude->SetIsGenius(Random(0.0f, 100.0f) > 60.0f);

                // Initialize him with the faction properties.
                dude->SetHue(m_factions[faction].hue);
                dude->SetCooperation(m_factions[faction].cooperation);

                m_dudes.push_back(dude);

                // Create a new alert where we spawned this guy.

                Alert* alert = new Alert;

                Color color  = GetFactionColor(faction);
                Vec2  origin = GetPoint(dude->GetPosition(), 10.0f);
                alert->Initialize(Color(0.0f, 0.0f, 0.0f), origin, 0.07f, 1000.0f);
                
                AddEffect(alert);

            }

         
        }

        m_dudeSpawnTimer -= dudeSpawnInterval;
    
    }

}

unsigned int Planet::GetRandomizedFactionToPopulate() const
{
    
    // Total the reproductive rate for all factions.

    float totalReproduction = 0.0f;

    for (unsigned int i = 0; i < m_factions.size(); ++i)
    {
        totalReproduction += m_factions[i].reproduction;
    }

    if (totalReproduction == 0.0f)
    {
        // This place is doomed!
        return -1;
    }

    float random = Random(0.0f, totalReproduction);

    for (unsigned int i = 0; i < m_factions.size(); ++i)
    {
        if (random <= m_factions[i].reproduction)
        {
            // Spawn in this faction.
            return i;
        }
        random -= m_factions[i].reproduction;
    }

    // No suitable factions.
    return -1;

}

Color Planet::GetFactionColor(unsigned int i) const
{
    Color color;
    color.SetHSV(m_factions[i].hue, 0.3f, 1.0f);
    return color;
}

void Planet::ProcessFactionDiplomacy()
{

    unsigned int prevFaction = GetNextOccupiedFaction(0);
    unsigned int faction     = GetNextOccupiedFaction(prevFaction);

    unsigned int start       = faction;
    
    do
    {

        unsigned int nextFaction = GetNextOccupiedFaction(faction);

        if (prevFaction != faction)
        {
            ProcessFactionDiplomacy(faction, prevFaction);
            if (prevFaction != nextFaction)
            {
                ProcessFactionDiplomacy(faction, nextFaction);
            }
        }
        
        prevFaction = faction;
        faction     = nextFaction;

    }
    while (faction != start);

}

void Planet::ProcessFactionDiplomacy(unsigned int factionIndex1, unsigned int factionIndex2)
{

    assert(m_factions[factionIndex1].numDudes > 0);
    assert(m_factions[factionIndex2].numDudes > 0);

    // Check if the other faction has as resource that we need.

    Faction& faction1 = m_factions[factionIndex1];
    Faction& faction2 = m_factions[factionIndex2];

    const float nukeCost = 20.0f;

    if (faction2.numDudes > faction1.numDudes)
    {
        if (faction1.knowledge > nukeCost)
        {

            // Nuke 'em!

            float xStart    = Random(m_factions[factionIndex1].min, m_factions[factionIndex1].max);
            float xEnd      = Random(m_factions[factionIndex2].min, m_factions[factionIndex2].max);

            Nuke* nuke = new Nuke;
            nuke->Initialize(xStart, xEnd, 0.0005f);

            AddEffect(nuke);

            DeductKnowledge(factionIndex1, nukeCost);
        
        }
    }

}

unsigned int Planet::GetNextOccupiedFaction(unsigned int faction) const
{
    
    unsigned int nextFaction = faction;
  
    do
    {
        nextFaction = (nextFaction + 1) % m_factions.size();
    }
    while (m_factions[nextFaction % m_factions.size()].numDudes == 0 && nextFaction != faction);

    return nextFaction;

}

void Planet::CreateExplosion(float x)
{

    // Cave in the landscape.
    RaiseLand(x, 0.05f, -25.0f);

    ParticleSystem* particleSystem;
    
    // Create a ring
    particleSystem = new ParticleSystem;
    particleSystem->InitializeCircle(Color(0.0f, 0.0f, 0.0f), GetPoint(x, 0.0f), 10.0f, 0.2f, 5000.0f, 200);
    AddEffect(particleSystem);

    // Create an inner core.
    particleSystem = new ParticleSystem;
    particleSystem->InitializeCircle(Color(0.0f, 0.0f, 0.0f), GetPoint(x, 0.0f), 20.0f, 0.1f, 5000.0f, 100);
    AddEffect(particleSystem);


}

void Planet::SpawnMeteors()
{

    const float meteorSpawnInterval = 30000.0f;

    while (m_meteorSpawnTimer < 0.0f)
    {

        Meteor* meteor = new Meteor;
        meteor->Initialize();

        m_meteors.push_back(meteor);
        m_meteorSpawnTimer += Random(0.0f, meteorSpawnInterval);

    }

}

void Planet::RenderMeteors() const
{
    for (unsigned int i = 0; i < m_meteors.size(); ++i)
    {
        m_meteors[i]->Render(*this);
    }
}

void Planet::RenderMeteorites() const
{
    for (unsigned int i = 0; i < m_meteorites.size(); ++i)
    {
        RenderThing(m_meteorites[i]);
    }
}

void Planet::UpdateMeteors(float deltaTime)
{
    
    for (unsigned int i = 0; i < m_meteors.size(); /* */)
    {

        m_meteors[i]->Update(*this, deltaTime);

        // Remove the dead guy.

        if (!m_meteors[i]->GetIsAlive())
        {
            delete m_meteors[i];
            m_meteors[i] = m_meteors[m_meteors.size() - 1];
            m_meteors.resize(m_meteors.size() - 1);
        }
        else
        {
            ++i;
        }

    }

}

void Planet::UpdateMeteorites(float deltaTime)
{
    
    for (unsigned int i = 0; i < m_meteorites.size(); /* */)
    {

        m_meteorites[i]->Update(*this, deltaTime);

        // Remove the dead guy.

        if (!m_meteorites[i]->GetIsAlive())
        {
            delete m_meteorites[i];
            m_meteorites[i] = m_meteorites[m_meteorites.size() - 1];
            m_meteorites.resize(m_meteorites.size() - 1);
        }
        else
        {
            ++i;
        }

    }

}

void Planet::AddMeteorite(Meteorite* meteorite)
{
    m_meteorites.push_back(meteorite);
}

void Planet::MakeFactionPurchases()
{

    const int lumberCost = 4;
    const int oreCost    = 1;

    for (unsigned int i = 0; i < m_factions.size(); ++i)
    {

        Faction& faction = m_factions[i];

        // Try to build a rocket.

        if (faction.numDudes > 0 && faction.knowledge > 50.0f)
        {

            // Create a rocket.
            Rocket* rocket = new Rocket;
            rocket->SetPosition(Random(faction.min, faction.max));
            m_rockets.push_back(rocket);

            DeductKnowledge(i, 50.0f);

        }
        else if (faction.numDudes > 0 && faction.cooperation > 0.25f && faction.numBuildings < 3)
        {

            // If we have enough lumber and ore for the next level of technology, 

            if (faction.numOre >= oreCost && faction.numPlants >= lumberCost)
            {

                float x;
                bool success = false;

                for (int test = 0; test < 20 && !success; ++test)
                {
                    x = Random(faction.min, faction.max);

                    if (!GetIsBuildingNear(x, 0.01f))
                    {
                        success = true;
                    }
                }

                if (success)
                {

                    ConsumeFactionOre(faction, oreCost);
                    faction.numOre -= oreCost;

                    ConsumeFactionLumber(faction, lumberCost);
                    faction.numPlants -= lumberCost;

                    // Create a building.

                    Building* building = new Building;
                    building->SetPosition(x);
                    m_buildings.push_back(building);

                }

            }
            

        }

    }

}

void Planet::ConsumeFactionOre(const Faction& faction, int cost)
{

    // Horrible, but works
    while (cost > 0)
    {

        int offset = rand();

        for (unsigned int i = 0; i < m_meteorites.size(); ++i)
        {
            unsigned int j = (i + offset) % m_meteorites.size();
            if (faction.ContainsPosition(m_meteorites[j]->GetPosition()))
            {
                m_meteorites[j]->Kill(*this);
                break;
            }
        }

        --cost;
   
    }

}

void Planet::ConsumeFactionLumber(const Faction& faction, int cost)
{

    // Horrible, but works
    while (cost > 0)
    {

        int offset = rand();

        for (unsigned int i = 0; i < m_plants.size(); ++i)
        {
            unsigned int j = (i + offset) % m_plants.size();
            if (faction.ContainsPosition(m_plants[j]->GetPosition()))
            {
                m_plants[j]->Kill(*this);
                break;
            }
        }

        --cost;
   
    }

}

void Planet::DeductKnowledge(unsigned int faction, float amount)
{

    float amountPerDude = amount / m_factions[faction].numDudes;

    for (unsigned int i = 0; i < m_dudes.size(); ++i)
    {
        if (m_dudes[i]->GetFaction() == faction)
        {
            float knowledge = m_dudes[i]->GetKnowledge();
            m_dudes[i]->SetKnowledge(knowledge - amount);
        }
    }

}

bool Planet::GetHasWon() const
{
    return m_won;
}