#ifndef PLANET_H
#define PLANET_H

#include "Vec2.h"
#include "Direction.h"
#include "Font.h"

#include <vector>

//
// Forward declarations.
//

class Dude;
class Effect;
class Plant;
class Thing;
class Meteor;
class Meteorite;
class Building;
class Rocket;

/**
 * Planet class. Stores the representation of the planet along with all of the
 * objects (dudes, effects, etc.) on it.
 */
class Planet
{

public:

    virtual ~Planet();

    void    Initialize();

    void    Render() const;
    void    Update(float deltaTime);

    /**
     * Renders a pop up (i.e. tool tip) for the specified location. If there's
     * nothing interesting under that location, no tool tip is rendered.
     */
    void RenderPopup(const Vec2& screenLocation, const Vec2& worldLocation) const;
    
    void    SetAngle(float angle);
    float   GetAngle() const;

    float   GetSurfaceHeight() const;

    void    RaiseLand(float x, float radius, float amount);

    /**
     * Gets a position in world space on the planet at location x and a
     * distance height above the ground.
     */
    Vec2 GetPoint(float x, float height) const;

    float GetLandHeight(float x) const;
    float GetLandSlope(float x, Direction direction) const;

    static float Fractal(float x);

    /**
     * Returns the slope where terrain is considered impasssable.
     */
    float GetMaxImpassableSlope() const;

    void AddEffect(Effect* effect);

    /**
     * Returns true if the point is underwater.
     */
    bool GetIsUnderwater(float x) const;

    /**
     * Returns the number of dudes currently alive in the world.
     */
    unsigned int GetNumDudes() const;

    /**
     * Returns true if the location can support a tree growing there.
     */
    bool GetCanSupportPlants(float x) const;

    /**
     * Returns true if the point is on the sunny side of the planet.
     */
    bool GetIsSunny(float x) const;

    /**
     * Returns the color that represents the ith faction on the planet.
     */
    Color GetFactionColor(unsigned int i) const;

    /**
     * Creates an explosion at the point x.
     */
    void CreateExplosion(float x);

    /**
     * Adds a stationary meteorite to the planet.
     */
    void AddMeteorite(Meteorite* meteorite);

    bool GetHasWon() const;

private:

    struct Faction
    {

        bool ContainsPosition(float x) const;

        float           min;
        float           max;
        float           hue;

        unsigned int    numDudes;
        unsigned int    numPlants;
        unsigned int    numOre;
        unsigned int    numBuildings;
        unsigned int    numGeniuses;
        unsigned int    numRockets;

        float           cooperation;
        float           knowledge;
        float           reproduction;

    };

    void UpdateLandSlope();

    void BuildFactions();

    void RenderWater() const;
    void RenderDudes() const;
    void RenderBuildings() const;
    void RenderPlants() const;
    void RenderFactions() const;
    void RenderSun() const;
    void RenderMeteors() const;
    void RenderMeteorites() const;
    void RenderRockets() const;

    void UpdateDudes(float deltaTime);
    void UpdatePlants(float deltaTime);
    void UpdateEffects(float deltaTime);
    void UpdateMeteorites(float deltaTime);
    void UpdateMeteors(float deltaTime);
    void UpdateBuildings(float deltaTime);

    void ConsumeFactionOre(const Faction& faction, int cost);
    void ConsumeFactionLumber(const Faction& faction, int cost);

    void MakeFactionPurchases();

    void Populate(unsigned int numDudes);

    float GetLandSlope(float x, float lookAhead) const;

    void ApplyFactions(float deltaTime);

    // Updates the faction for each dude based on his position.
    void UpdateDudeFactions();

    /**
     * Updates the stats for a faction (number of dudes, etc.) based on the
     * dudes currently in it. UpdateDudeFactions must be called before this.
     */
    void UpdateFactionStats();

    /**
     * Counts the number of resources in each faction.
     */   
    void UpdateFactionResourceCounts();

    /**
     * Returns true if the location is impassable by dudes (making it a faction
     * boundary).
     */
    bool GetIsImpassableBoundary(int i) const;

    /**
     * Gets the height of waves at a specified angle on the planet.
     */
    float GetWaveHeight(float angle) const;

    /**
     * Renders a popup box at the specified location.
     */
    void RenderPopup(const Vec2& screenLocation, const char* message, float xSize, float ySize) const;

    /**
     * Spawns plants on the surface of the planet where conditions are good for
     * growing.
     */
    void SpawnPlants();

    /**
     * Spawns dudes if the population is below the maximum.
     */
    void SpawnDudes();

    /**
     * Spawns meteors that fly in from outer space.
     */
    void SpawnMeteors();
    
    /**
     * Returns true if there's a plant within the radius of the point specified.
     */
    bool GetIsPlantNear(float x, float radius) const;

    /**
     * Renders a thing.
     */
    void RenderThing(const Thing* thing) const;

    /**
     * Randomly selects a faction to populate weighted by the reproduction
     * factor. If no suitable faction is found, the method return -1.
     */
    unsigned int GetRandomizedFactionToPopulate() const;

    /**
     * Inovokes diplomatic (or maybe not diplomatic) actions for the factions based
     * on the current state of the other factions.
     */
    void ProcessFactionDiplomacy();

    /**
     * Returns the index of the next occupied faction.
     */
    unsigned int GetNextOccupiedFaction(unsigned int faction) const;

    /**
     * Inovokes diplomatic (or maybe not diplomatic) actions for the faction1
     * against faction2.
     */
    void ProcessFactionDiplomacy(unsigned int factionIndex1, unsigned int factionIndex2);

    bool GetIsBuildingNear(float x, float radius) const;

    void DeductKnowledge(unsigned int faction, float amount);

private:

    static const int            s_numLandSamples         = 512;
    static const unsigned int   s_factionAnarchySize     = 80;     
    static const float          s_maxImpassableSlope;

    float                       m_angle;

    float                       m_landSlope[s_numLandSamples][2];
    float                       m_landHeight[s_numLandSamples];

    std::vector<Dude*>          m_dudes;
    std::vector<Faction>        m_factions;
    std::vector<Effect*>        m_effects;
    std::vector<Plant*>         m_plants;
    std::vector<Meteor*>        m_meteors;
    std::vector<Meteorite*>     m_meteorites;
    std::vector<Building*>      m_buildings;
    std::vector<Rocket*>        m_rockets;

    float                       m_waterLevel;
    float                       m_waterTime;

    Vec2                        m_sunCenter;
    float                       m_sunRadius;

    Font                        m_popupFont;

    float                       m_plantSpawnTimer;
    float                       m_dudeSpawnTimer;
    float                       m_meteorSpawnTimer;

    unsigned int                m_maxDudes;

    bool                        m_won;

};

#endif